from .process.handler import PreprocessHandler
from .train.handler import ModelHandler
from .essentials.handler import EssentialsHandler
from .graph import Graph
from ..util import get_typed_params


def generate_file(graph, target_column, task_type, colab=False):
    g = Graph(len(graph["nodes"]))
    graph_dict = {}
    for i in graph["nodes"]:
        graph_dict[(i["data"]["name"])] = i
    map = {}
    j = 0
    while j < (len(graph["nodes"])):
        for i in graph["nodes"]:
            map[(i["id"])] = j
            j += 1
    map2 = {y: x for x, y in map.items()}
    for i in graph["edges"]:
        g.addEdge(map[i["source"]], map[i["target"]])
    res = g.topologicalSort()
    steps = []
    for i in res:
        steps.append(map2[i])
    steps_list = {}
    for i in steps:
        for j in range(len(graph["nodes"])):
            if i == graph["nodes"][j]["id"]:
                steps_list[(graph["nodes"][j]["data"]["name"])] = graph["nodes"][j][
                    "data"
                ]["node_type"]
    if colab:
        output_file = "/content/output.py"
    else:
        output_file = "output.py"
    preprocess_handler = PreprocessHandler()
    train_handler = ModelHandler()
    essentials_handler = EssentialsHandler()
    with open(output_file, "w") as f:
        f.write("import huble\n")
        f.write("from huble import Dataset\n")
        f.write("def run_experiment(experiment):\n")
        for i in steps_list:
            node = graph_dict[i]
            params = get_typed_params(node["data"]["parameters"])
            if steps_list[i] == "preprocess":
                f.write(
                    "\t"
                    + preprocess_handler.return_function(
                        function_name=node["data"]["name"], params=params
                    )
                )
                f.write("\n")
            elif (
                steps_list[i] == "classification_model"
                or steps_list[i] == "regression_model"
                or steps_list[i] == "clustering_model"
                or steps_list[i] == "model"
            ):
                f.write(
                    "\t"
                    + train_handler.return_function(
                        function_name=node["data"]["name"], params=params
                    )
                )
                f.write("\n")
            elif steps_list[i] == "essential":
                f.write(
                    "\t"
                    + essentials_handler.return_function(
                        function_name=node["data"]["name"], params=params, task_type=task_type, target_column= target_column
                    )
                )
                f.write("\n")
            elif steps_list[i] == "evaluate_model":
                f.write(
                    # TODO: make changes for clustering
                    f"\tmetrics = huble.sklearn.evaluate_model(model=Model, training_data=training_dataset, test_dataset=test_dataset, target_column= '{target_column}', task_type='{task_type}' )"
                )
                f.write("\n")
            elif steps_list[i] == "primary_dataset":
                # TODO: Add support for other datasets
                f.write(f"\tdata = Dataset('{node['data']['url']}').dataframe\n")
        f.write("\texperiment.upload_metrics(metrics,input_format)")
        f.write("\n\tprint(\"Uploading model...\")")
        f.write("\n\texperiment.upload_model(filename)")



from boto3.session import Session

session = Session()
s3 = session.client(
    "s3",
    region_name="us-east-1",
    endpoint_url="https://s3.filebase.com",
    aws_access_key_id="709B6E02413B7282AC93",
    aws_secret_access_key="EyByFcqsBwTe4bWq6PwnoUJ9e2BqzShztJ48efVH",
)
# Upload the file to S3
s3.upload_file("tested.csv", "testsesaetsatast", "tested.csv")

# Get the response headers for the uploaded file
response = s3.head_object(Bucket="testsesaetsatast", Key="tested.csv")
headers = response.get("ResponseMetadata", {}).get("HTTPHeaders", {})

print(headers)



class EssentialsHandler:
    def __init__(self) -> None:
        pass

    def return_function(self, function_name: str, params: dict, task_type: str, target_column: str):

        essentials = {
            "Train-Test Split": self.__train_test_split,
            "Train Model": self.__train_model,
        }
        return essentials[function_name](params, task_type)

    def __train_test_split(self, params, target_column):
        parameters = {
            "test_size": params["test_size"],
        }
        return f"training_dataset, test_dataset, input_format = huble.sklearn.train_test_split(data=data,parameters={parameters}, target_column={target_column})"

    def __train_model(self, params, task_type):
        return f"Model, filename = huble.sklearn.train_model(data=training_dataset, model=model, column='{params['target_column']}', task_type='{task_type}')"



import evalml
import pandas as pd
from evalml.problem_types import detect_problem_type
from evalml.data_checks import DefaultDataChecks
from evalml.automl import AutoMLSearch
from evalml.pipelines.utils import generate_pipeline_code

def Automl(data, target_column):
    #split into X and y
    X = data.drop(target_column, axis=1)
    y = data[target_column]  

    #Predict task type
    y_task = pd.Series(y)
    task=(detect_problem_type(y_task))
    task_str = str(task)

    #train test split
    X_train, X_test, y_train, y_test = evalml.preprocessing.split_data(X, y, problem_type=task)

    #Datachecks
    objective=""
    if task_str =='binary':
      objective = objective+"log loss binary"
    elif task_str == 'multiclass':
      objective = objective+"log loss multiclass"
    elif task_str=='regression':
      objective = objective+"R2"
    data_checks = DefaultDataChecks(task, objective)
    messages = data_checks.validate(X_train, y_train)

    errors = [message for message in messages if message["level"] == "error"]
    warnings = [message for message in messages if message["level"] == "warning"]
    data_issues = [] 
    for warning in warnings:
        data_issues.append("Warning:" + warning["message"])

    for error in errors:
        data_issues.append("Error:" + error["message"])


    #AutoMl Search
    automl = AutoMLSearch(X_train=X_train, y_train=y_train, problem_type=task)
    automl.search()

    #Pipeline Rankings
    rankings = automl.rankings
    rankings_json = rankings.to_json(orient="table")

    #Best Pipeline
    best_pipeline = automl.best_pipeline
    
    steps = best_pipeline.parameters
    steps_list=[]
    for step in best_pipeline:
      steps_list.append(step)
  
    #metrics
    objectives=[]
    if task_str =='binary':
      objectives.extend(['mcc binary',
                        'log loss binary',
                        'gini',
                        'auc',
                        'recall',
                        'precision',
                        'f1',
                        'balanced accuracy binary',
                        'accuracy binary'])
    elif task_str == 'multiclass':
      objectives.extend(['mcc multiclass',
                        'log loss multiclass',
                        'auc weighted',
                        'auc macro',
                        'auc micro',
                        'recall weighted',
                        'recall macro',
                        'recall micro',
                        'precision weighted',
                        'precision macro',
                        'precision micro',
                        'f1 weighted',
                        'f1 macro',
                        'f1 micro',
                        'balanced accuracy multiclass',
                        'accuracy multiclass'])
    elif task_str=='regression':
      objectives.extend(['expvariance',
                        'maxerror',
                        'medianae',
                        'mse',
                        'mae',
                        'r2',
                        'mean squared log error',
                        'root mean squared log error',
                        'root mean squared error',
                        'mean absolute percentage error'])
    metrics = dict(best_pipeline.score(X_test, y_test, objectives)) 

    #Feature importance for best pipeline
    feature_imp = best_pipeline.feature_importance
    feature_imp_json = feature_imp.to_json(orient="table")

    #generating code for best_pipeline
    code = generate_pipeline_code(best_pipeline)

    return data_issues, task_str, rankings_json, steps_list, feature_imp_json, code, best_pipeline, metrics



import pandas as pd
from sklearn.graph import Graph
from collections import Counter

def verify_pipeline(graph):
    issues=[]
    steps= {}
    for node in graph['nodes']:
        steps[node['data']['name']] =node['data']['node_type']
        if node['data']['node_type']=='primary_dataset' or node['data']['node_type']=='model':
            if 'target' not in node:
                issues.append(node['data']['name'] + ' Node not connected')
        elif node['data']['node_type']=='evaluate_model':
            if 'source' not in node:
                issues.append(node['data']['name'] + ' Node not connected')
        else:
            if 'source' not in node or 'target' not in node:
                issues.append(node['data']['name'] + ' Node not connected')

    value_counter = Counter(steps.values())
    key_counter = Counter(steps.keys())

    if 'Train Model' not in steps:
      issues.append('Train Model Node Missing')
    elif key_counter['Train Model']>1:
      issues.append('More than one Train Model Node found')
    if 'Train-Test Split' not in steps:
      issues.append('Train-Test Split Node Missing')
    elif key_counter['Train-Test Split']>1:
      issues.append('More than one Train-Test Split Node found')
    if 'Evaluate Model' not in steps:
      issues.append('Evaluate Model Node Missing')
    elif key_counter['Evaluate Model']>1:
      issues.append('More than one Evaluate Model Node found')
    if 'model' not in steps.values():
      issues.append('Model Node Missing')
    elif value_counter['model']>1:
      issues.append('More than one Model Node found')
    if 'primary_dataset' not in steps.values():
      issues.append('Dataset Node Missing')
    elif value_counter['primary_dataset']>1:
      issues.append('More than one Dataset Node found')

    g = Graph(len(graph["nodes"]))
    graph_dict = {}
    for i in graph["nodes"]:
        graph_dict[(i["data"]["name"])] = i
    map = {}
    j = 0
    while j < (len(graph["nodes"])):
        for i in graph["nodes"]:
            map[(i["id"])] = j
            j += 1
    map2 = {y: x for x, y in map.items()}
    for i in graph["edges"]:
        g.addEdge(map[i["source"]], map[i["target"]])

    if g.isCyclic() == 1:
        issues.append("Graph contains cycle")
    else:
        pass

    return issues





from huble.sklearn.metrics import log_metrics


def evaluate_model(model, training_dataset, test_dataset, target_column, task_type):
    if task_type=="classification" or task_type=="regression":
        y_test = test_dataset[target_column]
        X_test = test_dataset.drop([target_column], axis=1)
        y_pred = model.predict(X_test)
        metrics = log_metrics(y_true=y_test, y_pred=y_pred, task=task_type, X=None, labels=None)
        return metrics
    elif task_type=="clustering":
        labels = model.labels_
        metrics = log_metrics(y_true=None, y_pred=None, task=task_type, X=training_dataset, labels=labels)
        return metrics




from .functions import (
    remove_nan_values,
    replace_nan_values,
    drop_rows_columns,
    remove_outliers,
    drop_duplicates,
    change_data_type,
    min_max_scale,
    max_abs_scale,
    robust_scale,
    standard_scale,
    normalize,
    ordinal_encode,
    one_hot_encode,
    remove_mismatch_data,
    clean_data,
    round_data,
    filter_dataframe,
    truncate_dataframe,
    sort_values,
    transpose,
    rename_columns,
    select_columns,
    clean_column_names,
    clip,
    merge,
    clean_data,
    split_data,
)


from .handler import *



def inference_pip(graph):
    id='0'
    for node in graph['nodes']:
        if node['data']['node_type']=='primary_dataset':
            node['data']['name']='Web API'
        elif node['data']['name']=='Train-Test Split':
            node['data']['name']='Predictions'
            node['data']['value']='Predictions' 
            node['data']['node_type']='predictions'
            node.pop('target')
            node['data'].pop('outputParameters')
        elif node['data']['name']=='Train Model':
            id = node['id']
        graph['edges'] = [edge for edge in graph['edges'] if edge['target'] not in ['998', id]]
        graph['nodes'] = [node for node in graph['nodes'] if node['data']['node_type'] in ['primary_dataset', 'preprocess', 'predictions']]

    return graph



from mlxtend.feature_selection import ExhaustiveFeatureSelector as EFS


def feature_selection(X, y, model):
    efs = EFS(
        estimator=model,
        min_features=1,
        max_features=len(X.columns),
        scoring="accuracy",
        cv=5,
    )
    efs = efs.fit(X, y)
    features = []
    for feature in efs.best_feature_names_:
        features.append(feature)
    return features



from .sklearn import generate_file
from .connector import Experiment


def generate_experiment(experiment_id: str, auth_key=""):
    # TODO: Implement auth keys
    experiment = Experiment(experiment_id=experiment_id)
    print(experiment.target_column)
    generate_file(
        graph=experiment.graph,
        task_type=experiment.task_type,
        target_column=experiment.target_column,
    )
    return experiment



from sklearn import metrics
import evalml
import pandas as pd
from evalml.problem_types import detect_problem_type

def regression_metrics(y_true, y_pred):
    max_error = metrics.max_error(y_true, y_pred)
    mean_absolute_error = metrics.mean_absolute_error(y_true, y_pred)
    mean_squared_error = metrics.mean_squared_error(y_true, y_pred)
    mean_squared_log_error = metrics.mean_squared_log_error(y_true, y_pred)
    median_absolute_error = metrics.median_absolute_error(y_true, y_pred)
    r2_score = metrics.r2_score(y_true, y_pred)
    mean_poisson_deviance = metrics.mean_poisson_deviance(y_true, y_pred)
    mean_gamma_deviance = metrics.mean_gamma_deviance(y_true, y_pred)
    mean_tweedie_deviance = metrics.mean_tweedie_deviance(y_true, y_pred)
    # create dictionary of metrics
    metrics_dict = {
        "max_error": max_error,
        "mean_absolute_error": mean_absolute_error,
        "mean_squared_error": mean_squared_error,
        "mean_squared_log_error": mean_squared_log_error,
        "median_absolute_error": median_absolute_error,
        "r2_score": r2_score,
        "mean_poisson_deviance": mean_poisson_deviance,
        "mean_gamma_deviance": mean_gamma_deviance,
        "mean_tweedie_deviance": mean_tweedie_deviance,
    }
    return metrics_dict


def classification_metrics(y_true, y_pred):
    accuracy_score = metrics.accuracy_score(y_true, y_pred)
    balanced_accuracy_score = metrics.balanced_accuracy_score(y_true, y_pred)
    cohen_kappa_score = metrics.cohen_kappa_score(y_true, y_pred)
    f1_score = metrics.f1_score(y_true, y_pred)
    fbeta_score = metrics.fbeta_score(y_true, y_pred, beta=0.5)
    hamming_loss = metrics.hamming_loss(y_true, y_pred)
    jaccard_score = metrics.jaccard_score(y_true, y_pred)
    log_loss = metrics.log_loss(y_true, y_pred)
    matthews_corrcoef = metrics.matthews_corrcoef(y_true, y_pred)
    precision_score = metrics.precision_score(y_true, y_pred)
    recall_score = metrics.recall_score(y_true, y_pred)
    zero_one_loss = metrics.zero_one_loss(y_true, y_pred)
    roc_auc_score = metrics.roc_auc_score(y_true, y_pred)
    metrics_dict={}
    y_task = pd.Series(y_true)
    task=(detect_problem_type(y_task))
    task_str = str(task)
    if task_str == 'binary':
        metrics_dict = {
            "accuracy_score": accuracy_score,
            "balanced_accuracy_score": balanced_accuracy_score,
            "precision_score": precision_score,
            "recall_score": recall_score,
            "f1_score": f1_score,
            "roc_auc_score": roc_auc_score,
            "log_loss": log_loss,
            "matthews_corrcoef": matthews_corrcoef,
        }
    elif task == 'multiclass':
        metrics_dict = {
            "accuracy_score": accuracy_score,
            "balanced_accuracy_score": balanced_accuracy_score,
            "precision_score": precision_score,
            "recall_score": recall_score,
            "hamming_loss": hamming_loss,
            "jaccard_score": jaccard_score,
            "log_loss": log_loss,
            "cohen_kappa_score": cohen_kappa_score,
            "f1_score": f1_score,           
            "zero_one_loss": zero_one_loss,
        }
    return metrics_dict


def clustering_metrics(X, labels):
    silhouette_score = metrics.silhouette_score(X, labels)
    calinski_harabasz_score = metrics.calinski_harabasz_score(X, labels)
    davies_bouldin_score = metrics.davies_bouldin_score(X, labels)
    
    # create dictionary of metrics
    metrics_dict = {
        "silhouette_score": silhouette_score,
        "calinski_harabasz_score": calinski_harabasz_score,
        "davies_bouldin_score": davies_bouldin_score,
    }
    return metrics_dict


def log_metrics(y_true, y_pred, task, X, labels):
    if task == "regression":
        metrics_dict = regression_metrics(y_true, y_pred)
        return metrics_dict
    elif task == "classification":
        metrics_dict = classification_metrics(y_true, y_pred)
        return metrics_dict
    elif task == "clustering":
        metrics_dict = clustering_metrics(X, labels)
        return metrics_dict



import huble
from huble import Dataset
def run_experiment(experiment):
	model = huble.sklearn.decision_tree(parameters={'criterion': 'gini', 'splitter': 'best', 'max_depth': None, 'max_leaf_nodes': None, 'random_state': None})
	data = Dataset('https://ipfs.filebase.io/ipfs/QmRspeqXi9J2PVTmXYwMaBif9dYWVkNhM8EFomUAfajnT1').dataframe
	data = huble.sklearn.clean_data(data=data)
	training_dataset, test_dataset = huble.sklearn.train_test_split(data=data,parameters={'test_size': 0.2})
	Model, input_format, filename = huble.sklearn.train_model(data=training_dataset, model=model, column='Survived', task_type='classification')
	metrics = huble.sklearn.evaluate_model(model=Model, training_data=training_dataset, test_dataset=test_dataset, target_column= 'Survived', task_type='classification' )
	experiment.upload_metrics(metrics,input_format)
	print("Uploading model...")
	experiment.upload_model(filename)


import requests
from boto3.session import Session
import json


class Experiment:
    def __init__(self, experiment_id: str) -> None:
        self.experiment_id = experiment_id
        self.graph, self.task_type, self.target_column,self.modelURL,self.input_format = self.__get_experiment_details(experiment_id=experiment_id)

    def __get_experiment_details(self, experiment_id: str):
        try:
            response = requests.get(
                f"http://localhost:8000/experiments/{experiment_id}",
            )
            response = response.json()
        except:
            try:
                response = requests.get(
                    f"http://const_server:8000/experiments/{experiment_id}",
                )
                response = response.json()
            except:
              raise Exception("Unable to connect to the server")
        try:
            graph = response["pipelineJSON"]
            project = response["project"]
            target_column = project["targetColumn"]
            task_type = project["taskType"]
            modelURL = response["modelURL"]
            input_format = response["inputFormat"]
        except:
            raise Exception("Invalid Experiment ID")
        return graph, task_type, target_column,modelURL,input_format

    def upload_metrics(self, metrics, input_format):
        requests.put(
            f"http://localhost:8000/experiments/results/{self.experiment_id}",
            data={"metrics": json.dumps(metrics), "input_format": json.dumps(input_format)},
        )

    def upload_model(self, file_name):
        session = Session()
        client = session.client(
            "s3",
            region_name="us-east-1",
            endpoint_url="https://s3.filebase.com",
            aws_access_key_id="709B6E02413B7282AC93",
            aws_secret_access_key="EyByFcqsBwTe4bWq6PwnoUJ9e2BqzShztJ48efVH",
        )
        bucket_name = "testsesaetsatast"
        with open(file_name, "rb") as data:
            client.upload_fileobj(data, bucket_name, file_name)
        response = client.head_object(Bucket=bucket_name, Key=file_name)
        headers = response.get("ResponseMetadata", {}).get("HTTPHeaders", {})
        cid = headers.get("x-amz-meta-cid")
        requests.put(
            f"http://localhost:8000/experiments/model/{self.experiment_id}",
            data={"model": cid},
        )




from huble.error.exceptions import FunctionRuntimeException


def function_error_handling(name: str):

    """
    Decorator for handling errors while calling functions in Huble.
    """

    def wrapper(func):
        def wrapped(**kwargs):

            try:
                func(**kwargs)
            except Exception as e:
                raise FunctionRuntimeException(
                    f"{name} failed to execute. Error: {e}. Parameters given: {kwargs}"
                )

        return wrapped

    return wrapper



def get_typed_params(params: dict) -> dict:

    for param in params:
        if params[param] == "True":
            params[param] = True
            continue
        elif params[param] == "False":
            params[param] = False
            continue
        elif params[param] == "None":
            params[param] = None
            continue
        elif type(params[param]) == list:
            for i in params[param]:
                i = str(i)
                continue
        elif type(params[param]) == tuple:
            for i in params[param]:
                i = float(i)
            continue
        if (
            type(params[param]) is not list
            and type(params[param]) is not tuple
            and type(params[param]) is not dict
            and type(params[param]) is not bool
            and params[param] is not None
        ):
            try:
                params[param] = int(params[param])
            except:
                try:
                    params[param] = float(params[param])
                except:
                    try:
                        params[param] = str(params[param])
                    except:
                        print("Error generating type for parameter: ", param)

    return params
